#!/usr/bin/env python3
"""
This example illustrates how to perform density estimation by transforming
the data sample and reconstructing the comparison density with LOrPE. The
idea of this method is to model nonparametrically the deviations of the
data from some smooth parametric density. The hope is that these deviations
could be relatively mild and could be easier to reconstruct than the
original distribution itself.
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Oct 02 2019"

# Perform necessary imports
import math
import npstat as ns
from npstat_utils import HistoND, SampleMLFunctor1D
from npstat_density_estimation import LOrPE1D, histoPluginBandwidth1D, \
    BinnedRLCVCalc1D
import matplotlib.pyplot as plt
import scipy.optimize as opt

# Some parameters for the script
sampleSize = 500     # Number of points to generate (sample size)
xmin = -3.0          # Left edge of the sample histogram
xmax = 4.0           # Right edge of the sample histogram
numBins = 35         # Number of bins for the sample histogram
nDiscrete = 350      # Number of bins for discretization
symbetaPower = 4     # Power parameter of the symmetric beta kernel

# We will use a 1-d Gassian mixture to generate the random sample.
# We will truncate it to [xmin, xmax] in order to limit the support,
# otherwise we will not be able to construct the true comparison density.
component1 = ns.GaussianMixtureEntry(7.0, -1.0, 0.7)
component2 = ns.GaussianMixtureEntry(3.0,  1.0, 1.0)
mixture0 = ns.GaussianMixture1D(0.0, 1.0, (component1, component2))
mixture = ns.TruncatedDistribution1D(mixture0, xmin, xmax)

# Generate a random sample according to this density
rng = ns.MersenneTwister()
sample = mixture.generate(rng, sampleSize)

# Histogram for visualizing the sample
h1 = HistoND("int", ns.HistoAxis(numBins, xmin, xmax, "X"),
             "Sample Histogram", "Counts")
h1.fill(sample, 1)

# We will use a Beta distribution to provide an initial guess for the
# density estimation procedure. Below, the parameters of the beta are
# fitted to the sample data by maximum likelihood. To do this, we need
# to come up with a proper functor for "scipy.optimize". It is convenient
# to derive such a functor from the "SampleMLFunctor1D" class. This works
# for unbinned data. For fitting histograms, there is another functor
# which serves similar purpose: "HistoMLFunctor1D". In either case, we
# need to override the method "makeDistro" which bulids a distribution
# (an instance of the NPStat AbsDistribution1D class) out of the fitted
# parameters.
class Beta1DFunctor(SampleMLFunctor1D):
    def __init__(self, sample, xmin, xmax):
        SampleMLFunctor1D.__init__(self, sample)
        self._xmin = xmin
        self._width = xmax - xmin
    def makeDistro(self, params):
        # Extend the support of the Beta just a little bit, so that the
        # comparison density does not blow to infinity at xmin and xmax
        b = ns.Beta1D(self._xmin-0.01*self._width, self._width*1.02, *params)
        return ns.TruncatedDistribution1D(b, xmin, xmax)

# An instance of the functor
betaFcn = Beta1DFunctor(sample, xmin, xmax)

# Initial guess of the Beta distribution parameters.
# This guess matches sample mean and standard deviation.
mean, sigma, skewness, kurtosis = ns.arrayStats(sample)
m01 = (mean - xmin)/(xmax - xmin)
var01 = math.pow(sigma/(xmax - xmin), 2)
alpha0 = m01*(m01 - m01*m01 - var01)/var01
beta0 = (m01 - 1)*(m01*m01 + var01 - m01)/var01

# Minimize negative log-likelihood
minimum = opt.minimize(betaFcn, (alpha0, beta0),
                       method='L-BFGS-B', options = {'eps': 1e-5})
assert minimum.success, "Minimization did not converge"

# Transform the sample to [0, 1] using the CDF of the fitted density
beta = betaFcn.makeDistro(minimum.x)
transformedSample = [beta.cdf(x) for x in sample]

# Histogram for discretizing the transformed sample
h2 = HistoND("int", ns.HistoAxis(nDiscrete, 0.0, 1.0, "CDF"),
             "Comparison Density", "Counts")
h2.fill(transformedSample, 1)

# Fit the transformed sample histogram using LOrPE
smoother = LOrPE1D(h2)

# Add a number of filters with different degrees and bandwidth
# settings for cross-validation calculations. The logic of
# automatic bandwidth selection here is not foolproof -- we
# don't really know how well the initial fit matches the sample.
# If the match is perfect then we should just use the constant
# fit (or a very wide bandwidth) instead of applying the usual
# plug-in guess to initialize the cross-validation.
for filterDegree in range(1, 7):
    bw0 = histoPluginBandwidth1D(h2, sampleSize, filterDegree, symbetaPower)
    print("Building degree {} filters, plug-in bandwidth is {:.4f}".format(
        filterDegree, bw0))
    for bw in ns.EquidistantInLogSpace(bw0/2.0, bw0*4.0, 40):
        smoother.addFilter((filterDegree, bw, symbetaPower))

# Run the regularized likelihood cross-validation (RLCV)
regularisationParameter = 0.5
rlcvCalc = BinnedRLCVCalc1D(regularisationParameter)
bestRLCV, paramsRLCV, yRLCV = smoother.bestFilterByCV(h2, rlcvCalc)
rlcvDegree, rlcvBandwidth, rlcvSymbetaPower = paramsRLCV
print("RLCV: bandwidth is {:.4f}, degree is {}".format(rlcvBandwidth, rlcvDegree))

# Construct the composite distribution
comparisonDensity = ns.BinnedDensity1D(0.0, 1.0, yRLCV, 1)
fitted = ns.CompositeDistribution1D(comparisonDensity, beta)

# We are done with density estimation at this point. The rest of the script
# is just plotting results.
fig, axs = plt.subplots(1, 2)

# Plot the histogram of the transformed sample
h3 = HistoND("int", ns.HistoAxis(numBins, 0.0, 1.0, "CDF"),
             "Comparison Density", "Counts")
h3.fill(transformedSample, 1)
xh, yh = ns.histoOutline1D(h3)
axs[0].plot(xh, yh, 'k')

# Overlay the true comparison distribution
coords = h2.axis(0).binCenters()
verticalScale = h3.binVolume()*sampleSize
trueCD = ns.ComparisonDistribution1D(mixture, beta)
yc = [trueCD.density(x)*verticalScale for x in coords]
axs[0].plot(coords, yc, 'b', label='True density')

# Overlay the LOrPE RLCV result
axs[0].plot(coords, yRLCV*verticalScale, 'r', label='LOrPE RLCV')

# Set labels and title for this panel
axs[0].set_xlabel(h3.axis(0).label())
axs[0].set_ylabel('Density estimate')
axs[0].set_title('Comparison Density Fit')
axs[0].legend(loc=1)

# Plot the sample histogram
xh, yh = ns.histoOutline1D(h1)
axs[1].plot(xh, yh, 'k')

# Overlay the original density. Scale all densities so that their
# plots are compatible with the histogram height.
verticalScale = h1.binVolume()*sampleSize
xdens, ydens = ns.scanDensity1D(mixture, xmin, xmax, nDiscrete+1)
axs[1].plot(xdens, ydens*verticalScale, 'b', label='True density')

# Overlay the initial fit of the Beta function
x0, y0 = ns.scanDensity1D(beta, xmin, xmax, nDiscrete+1)
axs[1].plot(x0, y0*verticalScale, 'g', label='Initial Fit')

# Overlay the LOrPE result
xlorpe, ylorpe = ns.scanDensity1D(fitted, xmin, xmax, nDiscrete+1)
axs[1].plot(xlorpe, ylorpe*verticalScale, 'r', label='LOrPE')

# Add labels and comments
axs[1].set_xlabel(h1.axis(0).label())
axs[1].set_ylabel('Density estimate')
axs[1].set_title('Overall Fit')
axs[1].legend(loc=1)

# Display the window
fig.tight_layout()
plt.show()
