#!/usr/bin/env python3
"""
This example illustrates how to choose 1-d LOrPE filter settings
by cross-validation
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Sep 19 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
from npstat_density_estimation import LOrPE1D, FFTKDE1D, \
    histoPluginBandwidth1D, BinnedLSCVCalc1D, BinnedRLCVCalc1D
import matplotlib.pyplot as plt
from npstat_plotting import plotHistoOutline1D

# Some parameters for the script
sampleSize = 1000    # Number of points to generate (sample size)
xmin = 0.0           # Left edge of the sample histogram
xmax = 5.0           # Right edge of the sample histogram
numBins = 100        # Number of bins for the sample histogram
nDiscrete = 500      # Number of cells for sample discretization.
                     # Should be large enough so that the cell width
                     # is substantially smaller than any physical scale.
symbetaPower = 4     # Power parameter of the symmetric beta kernel.
                     # Specify -1 to use a Gaussian kernel instead.
pluginDegree = 2     # Degree of polynomial used to generate a simple
                     # plug-in estimate for comparison with cross-validation
allowNegative = True # Specify whether the density estimate is allowed
                     # to be negative somewhere (this is possible in case
                     # filterDegree > 0)
bh = "BM_FOLD"       # LOrPE boundary handling method. See "lorpe_1d.py"
                     # script for more details.

# Here, we will use the exponential distribution truncated to [xmin, xmax]
# in order to generate the random sample
d1 = ns.TruncatedDistribution1D(ns.Exponential1D(0.0, 1.0), xmin, xmax)

# Generate a random sample according to this density
rng = ns.MersenneTwister()
sample = d1.generate(rng, sampleSize)

# Histogram for displaying the sample
h1 = HistoND("int", ns.HistoAxis(numBins, xmin, xmax, "X"),
             "Sample Histogram", "Counts")
h1.fill(sample, 1)

# Histogram for sample discretization
h2 = HistoND("int", ns.HistoAxis(nDiscrete, xmin, xmax, "X"),
             "LOrPE Discretization Histogram", "Counts")
h2.fill(sample, 1)

# Object which will perform LOrPE
smoother = LOrPE1D(h2, allowNegative, ns.BoundaryHandling(bh))

# Add a density estimate with a simple plug-in bandwidth
bw0 = histoPluginBandwidth1D(h2, sampleSize, pluginDegree, symbetaPower)
print("Plug-in bandwidth is {:.4f} for degree {}".format(bw0, pluginDegree))
yPlugin = smoother.estimateDensity(h2, (pluginDegree, bw0, symbetaPower))

# Add vanilla KDE for comparison (no mirroring and no high order kernels)
kde = FFTKDE1D(h2, False, False)
bwkde = histoPluginBandwidth1D(h2, sampleSize, 0, symbetaPower)
yKDE = kde.estimateDensity(h2, (0, bwkde, symbetaPower))
print("KDE bandwidth is {:.4f}".format(bwkde))

# Add a number of filters with different degrees and bandwidth
# settings for cross-validation calculations
for filterDegree in range(1, 7):
    bw0 = histoPluginBandwidth1D(h2, sampleSize, filterDegree, symbetaPower)
    print("Building degree {} filters, plug-in bandwidth is {:.4f}".format(
        filterDegree, bw0))
    for bw in ns.EquidistantInLogSpace(bw0/4.0, bw0*4.0, 50):
        smoother.addFilter((filterDegree, bw, symbetaPower))

# Run the least squares cross-validation (LSCV)
lscvCalc = BinnedLSCVCalc1D()
bestLSCV, paramsLSCV, yLSCV = smoother.bestFilterByCV(h2, lscvCalc)
lscvDegree, lscvBandwidth, lscvSymbetaPower = paramsLSCV
print("LSCV: bandwidth is {:.4f}, degree is {}".format(lscvBandwidth, lscvDegree))

# Run the regularized likelihood cross-validation (RLCV)
regularisationParameter = 0.5
rlcvCalc = BinnedRLCVCalc1D(regularisationParameter)
bestRLCV, paramsRLCV, yRLCV = smoother.bestFilterByCV(h2, rlcvCalc)
rlcvDegree, rlcvBandwidth, rlcvSymbetaPower = paramsRLCV
print("RLCV: bandwidth is {:.4f}, degree is {}".format(rlcvBandwidth, rlcvDegree))

# Create Matplotlib plot window
fig = plt.figure()
ax = fig.add_subplot(111)

# Plot the sample histogram
plotHistoOutline1D(ax, h1, ylabel='Density estimate',
                   title='LOrPE with Cross-Validation')

# Overlay the original density. Scale all densities so that their
# plots are compatible with the histogram height.
verticalScale = h1.binVolume()*sampleSize
xdens, ydens = ns.scanDensity1D(d1, xmin, xmax, nDiscrete+1)
ax.plot(xdens, ydens*verticalScale, 'g', label='True density')

# Overlay the KDE
coords = h2.axis(0).binCenters()
ax.plot(coords, yKDE*verticalScale, 'm', label="KDE")

# Overlay the LOrPE results with various bandwidth values
colors     = ('c',        'r',      'b'   )
labels     = ('Plug-in',  'RLCV',   'LSCV')
results    = (yPlugin,    yRLCV,    yLSCV )
for color, l, y in zip(colors, labels, results):
    ax.plot(coords, y*verticalScale, color, label="{} estimate".format(l))

# Add plot legend
plt.legend()

# Display the window
plt.show()
