#!/usr/bin/env python3
"""
This example illustrates how to choose multivariate LOrPE filter
settings by cross-validation
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Sep 30 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
from npstat_density_estimation import LOrPEND, FFTKDEND, \
    normalBandwidthFactor, scanMultivariateDensityAsWeight, \
    BinnedRLCVCalcND, BinnedLSCVCalcND
import matplotlib.pyplot as plt
from npstat_plotting import *

# Some parameters for the script
sampleSize = 1000    # Number of points to generate (sample size)
nDiscrete = 70       # Number of discretization cells to use (here, in each
                     # dimension). Should be a power of 2 for best FFT
                     # performance.
filterDegree = 0     # Polynomial degree used to generate high-order kernels.
                     # In general, even numbers make the most sense here.
                     # Kernel order will be filterDegree + 2.
allowNegative = True # Specify whether the density estimate is allowed
                     # to be negative somewhere (this is possible in case
                     # filterDegree > 0)

# We will use a 2-d Gassian mixture to generate the random sample.
# The code below constructs such a mixture. The weights do not have
# to add up to 1, they will be renormalized internally.
mixture = ns.DistributionMixND(2)
center0 = (1.0, 0.5)
covarianceMatrix0 = ((2.0, 0.3),
                     (0.3, 0.5))
weight0 = 7
mixture.add(ns.GaussND(center0, covarianceMatrix0), weight0)
center1 = (-1.0, -0.5)
covarianceMatrix1 = ((2.0, -2.0),
                    (-2.0,  3.0))
weight1 = 3
mixture.add(ns.GaussND(center1, covarianceMatrix1), weight1)

# Generate a random sample according to this density.
# Note that the "generate" function of multivariate densities
# generates a flat list of doubles, not a multivariate sequence.
# The normal "fill" method of the HistoND class will work
# in such cases as well, consuming several coordinate values
# at a time. The number of elelements in the list should be
# divisible by the histogram dimensionality.
rng = ns.MersenneTwister()
sample = mixture.generate(rng, sampleSize)

# Histogram for visualizing the sample
xaxis = ns.HistoAxis(25, -5.0, 5.0, "X")
yaxis = ns.HistoAxis(25, -5.0, 5.0, "Y")
h1 = HistoND("int", xaxis, yaxis, "Sample Histogram", "Counts")
h1.fill(sample, 1)

# Histogram for sample discretization. Note that the axis range is
# increased, to catch all possible tails. In general, such an increase
# may or may not make sense, depending on situation.
h2 = HistoND("int", ns.HistoAxis(nDiscrete, -7.0, 7.0, "X"),
                    ns.HistoAxis(nDiscrete, -7.0, 7.0, "Y"),
                    "KDE Discretization Histogram", "Counts")
h2.fill(sample, 1)

# Object which will perform the KDE
smoother = LOrPEND(h2, allowNegative)
smoother2 = FFTKDEND(h2, allowNegative)

# Make an initial guess for the filter bandwidth. This guess will use
# the sample covariance matrix estimated from the histogram. Ideally,
# here we need some robust covariance matrix estimate instead -- minimum
# volume ellipsoid (MVE), minimum covariance determinant (MCD), etc.
pluginFactor = normalBandwidthFactor(h2.dim(), filterDegree, sampleSize)
pluginCovariance = pluginFactor*ns.histoCovariance(h2)

# Build a kernel for the plug-in estimate
kernelDistro = ns.GaussND((0, 0), pluginCovariance)

# Discretize this kernel so that we can subsequently
# feed it to the smoother
kernelScan = scanMultivariateDensityAsWeight(kernelDistro, (1, 1), h2)

# Build the density estimate using the plug-in kernel
pluginEstimate = smoother.estimateDensity(h2, 1.0, filterDegree,
                                          kernelScan, False, False)
pluginEstimate2 = smoother2.estimateDensity(h2, 1.0, filterDegree,
                                            kernelScan, False, False)

# Add a number of kernels with different bandwidth values
# for cross-validation calculations. In general, filterDegree
# parameter can be simultaneously optimized as well.
for i, bwFactor in enumerate(ns.EquidistantInLogSpace(0.5, 2.0, 10)):
    print("Computing cross-validation filter", i)
    covMat = bwFactor*bwFactor*pluginCovariance
    kernelDistro = ns.GaussND((0, 0), covMat)
    kernelScan = scanMultivariateDensityAsWeight(kernelDistro, (1, 1), h2)
    smoother.addFilter(bwFactor, filterDegree, kernelScan, False)

# Run the regularized likelihood cross-validation (RLCV).
# For the description of this method, see section 4.2 in
# https://doi.org/10.1080/10485252.2017.1371715.
#
# Note that it takes a lot longer to compute multivariate LOrPE
# filters than to apply them, so it makes sense to compute the
# filters once and then use them to reconstruct multiple similar
# datasets (e.g., when you resample your data).
regularisationParameter = 0.5
rlcvCalc = BinnedRLCVCalcND(regularisationParameter)
bestRLCV, paramsRLCV, scanRLCV = smoother.bestFilterByCV(h2, rlcvCalc)
bwFactorRLCV, filterDegreeRLCV = paramsRLCV
print("Best RLCV bandwidth factor is {:.4f}".format(bwFactorRLCV))

# Calculate the integrated squared error (ISE) for various estimates
box = h2.boundingBox()
trueDensity = ns.scanDensityND(mixture, box, h2.shape())
for e, tag in zip((pluginEstimate, scanRLCV, pluginEstimate2),
                  ("Plug-in",      "RLCV",       "KDE")):
    ise = ns.sumOfSquaredDifferences(trueDensity, e)*h2.binVolume()
    print(tag, "estimate ISE is {:.5f}".format(ise))

# Plot various things
fig, axs = plt.subplots(2, 3)
colormeshNumpyArray(axs[0, 0], trueDensity, box,
                    flipy=True, title="True Density")
colormeshHisto2D(axs[0, 1], h1)
colormeshNumpyArray(axs[0, 2], pluginEstimate2, box,
                    flipy=True, title="KDE Plug-in")
colormeshNumpyArray(axs[1, 0], pluginEstimate, box,
                    flipy=True, title="LOrPE Plug-in")
colormeshNumpyArray(axs[1, 1], scanRLCV, box,
                    flipy=True, title="LOrPE RLCV")
colormeshNumpyArray(axs[1, 2], scanRLCV - trueDensity, box,
                    flipy=True, title="RLCV - True")

# Set x and y ranges of the plots to that of the sample histogram
for ax in [axs[i, j] for i in range(2) for j in range(3)]:
    ax.set_xlim(h1.axis(0).min(), h1.axis(0).max())
    ax.set_ylim(h1.axis(1).min(), h1.axis(1).max())

# Display the window
fig.suptitle("Multivariate LOrPE with Cross-Validation")
fig.tight_layout(rect=[0, 0, 1, 0.95])
plt.show()
