#!/usr/bin/env python3
"""
This example illustrates how to fill NPStat 2-d histograms and plot them
using Axes3D from Matplotlib
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Sep 13 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

# To construct an n-dimensional histogram, we need to provide
# a collection of n axes as the constructor argument.
xaxis = ns.HistoAxis(25, -5.0, 5.0, "X")
yaxis = ns.HistoAxis(10, -2.0, 3.0, "Y")

# The following line creates a 2-d histogram with bins of type "double".
# The histogram dimensionality is inferred from the number of axes provided.
h2 = HistoND("double", (xaxis, yaxis), "Example 2-d Histogram", "Z")

# We will fill the histogram with random numbers, so we need a generator
rng = ns.MersenneTwister()

# The flat random numbers on [0, 1) will be converted into multivariate
# Gaussian random numbers using an object of "GaussND" class.
center = (1.0, 0.5)
covarianceMatrix = ((1.0, 0.3),
                    (0.3, 0.25))
distro = ns.GaussND(center, covarianceMatrix)

# Fill the histogram. Here, each point contributes unit weight to
# the corresponding bin.
h2.fill(distro.generate(rng, 10000), 1.0)

# Prepare Matplotlib figure for plotting 3-d bars
fig = plt.figure()
ax = Axes3D(fig)

# Get the bars from the histogram using the NPStat "histoBars3d" function.
#
#  xscale is the scale factor for the bar width. Value of 1 corresponds
#         to the normal histogram bin width.
#
#  xshift is the shift of the bars with respect to their normal position
#         (centered in the corresponding bin slot), in the units of bin width
#
#  yscale, yshift -- same as xscale, xshift, but for the y direction
#
#  zscale is the vertical scale factor for the bars
#
#  zshift are the vertical shifts for the bars (in bin counter units, whatever
#         they are). The collection of z shifts can be empty in which case
#         it is assumed that all shifts are 0. If this collection is not
#         empty, its length must be equal to the number of histogram bins.
xscale = 0.8
xshift = 0.0
yscale = 0.8
yshift = 0.0
zscale = 1.0
zshift = []
x, y, z, dx, dy, dz = ns.histoBars3d(h2, xscale, xshift,
                                     yscale, yshift, zscale, zshift)

# Plot the bars
ax.bar3d(x, y, z, dx, dy, dz)

# Set the axis labels and the plot title
ax.set_xlabel(h2.axis(0).label())
ax.set_ylabel(h2.axis(1).label())
ax.set_zlabel(h2.accumulatedDataLabel())
plt.title(h2.title())

# Display the window. The code enters Matplotlib event loop here.
# You will be able to use the window interface to manipulate the figure
# (rotate, pan, zoom) and to save it in various graphical formats.
plt.show()
