#ifndef NPSTAT_SEMIINFGAUSSIANQUADRATURE_HH_
#define NPSTAT_SEMIINFGAUSSIANQUADRATURE_HH_

/*!
// \file SemiInfGaussianQuadrature.hh
//
// \brief Gaussian quadrature with weight exp(-x^2/2) on [0, Infinity]
//
// Author: I. Volobouev
//
// June 2019
*/

#include <vector>
#include <utility>

#include "npstat/nm/SimpleFunctors.hh"

namespace npstat {
    /** 
    // Quadrature for functions of the type f(x) exp(-x^2/2) on [0, Infinity]
    */
    class SemiInfGaussianQuadrature
    {
    public:
        /**
        // At the moment, the following numbers of points are supported:
        // 4, 8, 16, 24, 32.
        //
        // If an unsupported number of points is given in the
        // constructor, std::invalid_argument exception will be thrown.
        */
        explicit SemiInfGaussianQuadrature(unsigned npoints);

        /** Return the number of quadrature points */
        inline unsigned npoints() const {return npoints_;}

        /**
        // Get the abscissae for the integration points.
        // The buffer length should be at least npoints.
        */
        void getAbscissae(long double* abscissae, unsigned len) const;

        /**
        // Get the weights for the integration points.
        // The buffer length should be at least npoints.
        */
        void getWeights(long double* weights, unsigned len) const;

        /** Perform the quadrature */
        template <typename FcnResult, typename FcnArg>
        long double integrate(const Functor1<FcnResult,FcnArg>& fcn) const;

        /**
        // Perform the quadrature f(x) Sqrt[2/Pi]/sigma Exp[-(x/sigma)^2/2]
        // (the weight is normalized to unit integral on [0, Infinity])
        */
        template <typename FcnResult, typename FcnArg>
        long double integrateProb(long double sigma,
                                  const Functor1<FcnResult,FcnArg>& fcn) const;

        /** Check if the rule with the given number of points is supported */
        static bool isAllowed(unsigned npoints);

        /** The complete set of allowed rules, in the increasing order */
        static std::vector<unsigned> allowedNPonts();

        /**
        // Minimum number of points, among the supported rules, which
        // integrates a polynomial with the given degree exactly (with
        // the appropriate weight). Returns 0 if the degree is out of range.
        */
        static unsigned minimalExactRule(unsigned polyDegree);

    private:
        SemiInfGaussianQuadrature();

        const long double* a_;
        const long double* w_;
        mutable std::vector<std::pair<long double, long double> > buf_;
        unsigned npoints_;

#ifdef SWIG
    public:
        inline std::vector<double> abscissae2() const
        {
            return std::vector<double>(a_, a_+npoints_);
        }

        inline std::vector<double> weights2() const
        {
            return std::vector<double>(w_, w_+npoints_);
        }
#endif // SWIG
    };
}

#include "npstat/nm/SemiInfGaussianQuadrature.icc"

#endif // NPSTAT_SEMIINFGAUSSIANQUADRATURE_HH_
