#ifndef NPSTAT_DELTAMIXTURE1D_HH_
#define NPSTAT_DELTAMIXTURE1D_HH_

/*!
// \file DeltaMixture1D.hh
//
// \brief A mixture of 1-d Dirac delta functions
//
// Author: I. Volobouev
//
// July 2019
*/

#include <vector>

#include "npstat/stat/AbsDistribution1D.hh"
#include "npstat/stat/Distribution1DFactory.hh"

namespace npstat {
    class DeltaMixture1D : public AbsScalableDistribution1D
    {
    public:
        template <typename Real1, typename Real2>
        inline DeltaMixture1D(const double location, const double scale,
                              const Real1* unscaledCoords,
                              const Real2* unscaledWeights,
                              const unsigned arrayLength)
            : AbsScalableDistribution1D(location, scale)
        {initialize(unscaledCoords, unscaledWeights, arrayLength);}

        inline virtual DeltaMixture1D* clone() const
            {return new DeltaMixture1D(*this);}

        inline virtual ~DeltaMixture1D() {}

        /** Density integral on the intervals (a, b), [a, b], etc */
        double integral(double a, bool includePointAtA,
                        double b, bool includePointAtB) const;

        /** Distribution moment */
        long double moment(long double center, unsigned order) const;

        /** 
        // A collection of distribution moments. The array "moments"
        // should have at least "maxOrder+1" elements.
        */
        void moments(long double center, unsigned maxOrder,
                     long double* moments) const;

        /** 
        // A collection of distribution central moments. The array "moments"
        // should have at least "maxOrder+1" elements. On exit, moments[1]
        // will be set to the mean while all other moments will be central.
        */
        void centralMoments(unsigned maxOrder, long double* moments) const;

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream&) const;

        static inline const char* classname() {return "npstat::DeltaMixture1D";}
        static inline unsigned version() {return 1;}
        static DeltaMixture1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        friend class ScalableDistribution1DFactory<DeltaMixture1D>;

        DeltaMixture1D(double location, double scale,
                       const std::vector<double>& params);
        inline static int nParameters() {return -1;}

        template <typename Real1, typename Real2>
        void initialize(const Real1* unscaledCoords,
                        const Real2* unscaledWeights,
                        const unsigned arrayLength);
        void buildCdfTable();

        inline double unscaledDensity(double /* x */) const {return 0.0;}
        double unscaledCdf(double x) const;
        double unscaledExceedance(double x) const;
        double unscaledQuantile(double x) const;

        double weightSum(unsigned imin, unsigned imax) const;

        // Find all locations in the range
        void determineRange(double a, bool includeLeftBoundary,
                            double b, bool includeRightBoundary,
                            unsigned* imin, unsigned* imax) const;

        std::vector<double> locations_;
        std::vector<double> weights_;
        std::vector<double> cdf_;
        std::vector<double> exceedance_;
    };
}

#include "npstat/stat/DeltaMixture1D.icc"

#endif // NPSTAT_DELTAMIXTURE1D_HH_
