#ifndef NPSTAT_GAUSSIANMIXTUREENTRY_HH_
#define NPSTAT_GAUSSIANMIXTUREENTRY_HH_

/*!
// \file GaussianMixtureEntry.hh
//
// \brief A helper class for constructing one-dimensional Gaussian mixtures
//
// Author: I. Volobouev
//
// October 2011
*/

#include <stdexcept>

#include "geners/ClassId.hh"

namespace npstat {
    class GaussianMixture1D;

    /** A Gaussian mixture component */
    class GaussianMixtureEntry
    {
    public:
        inline GaussianMixtureEntry() : w_(0.0), mean_(0.0), stdev_(1.0) {}

        /** Constructor arguments are weight, mean, and standard deviation */
        inline GaussianMixtureEntry(const double w, const double mean,
                                    const double stdev)
            : w_(w), mean_(mean), stdev_(stdev)
        {
            if (w_ < 0.0) throw std::invalid_argument(
                "In npstat::GaussianMixtureEntry constructor: "
                "component weight must be non-negative");
            if (w_ > 0.0 && stdev_ <= 0.0) throw std::invalid_argument(
                "In npstat::GaussianMixtureEntry constructor: "
                "standard deviation must be positive");
        }

        //@{
        /** Examine object properties */
        inline double weight() const {return w_;}
        inline double mean() const {return mean_;}
        inline double stdev() const {return stdev_;}
        //@}

        bool operator==(const GaussianMixtureEntry& r) const;
        inline bool operator!=(const GaussianMixtureEntry& r) const
            {return !(*this == r);}

        //@{
        /** Method needed for "geners" I/O */
        inline gs::ClassId classId() const {return gs::ClassId(*this);}
        bool write(std::ostream& os) const;
        //@}

        static inline const char* classname()
            {return "npstat::GaussianMixtureEntry";}
        static inline unsigned version() {return 1;}
        static void restore(const gs::ClassId& id, std::istream& in,
                            GaussianMixtureEntry* entry);
    private:
        friend class GaussianMixture1D;

        double w_;
        double mean_;
        double stdev_;
    };
}

#endif // NPSTAT_GAUSSIANMIXTUREENTRY_HH_
