#include <cassert>
#include <stdexcept>
#include <cfloat>
#include <cmath>

#include "npstat/stat/correctDensityEstimateGHU.hh"

static double thresholdedIntegral(const double* arr, const unsigned long len,
                                  const double binWidth, const double threshold)
{
    long double sum = 0.0L;
    for (unsigned long i=0; i<len; ++i)
        if (arr[i] > threshold)
            sum += (arr[i] - threshold);
    return sum*binWidth;
}

namespace npstat {
    void correctDensityEstimateGHU(
        const double* estimator, const unsigned long len,
        double* corrected, const unsigned long lenCorrected,
        const double binWidth)
    {
        if (len != lenCorrected) throw std::invalid_argument(
            "In npstat::correctDensityEstimateGHU: incompatible buffers");
        if (len)
        {
            assert(estimator);
            assert(corrected);
            if (binWidth <= 0.0) throw std::invalid_argument(
                "In npstat::correctDensityEstimateGHU: bin size must be positive");

            bool haveNegative = false;
            long double sum = 0.0L;
            for (unsigned long i=0; i<len; ++i)
            {
                sum += estimator[i];
                if (estimator[i] < 0.0)
                    haveNegative = true;
            }
            const double integ = sum*binWidth;
            if (integ <= 0.0) throw std::invalid_argument(
                "In npstat::correctDensityEstimateGHU: "
                "estimator integral must be positive");
            for (unsigned long i=0; i<len; ++i)
                corrected[i] = estimator[i]/integ;
            if (haveNegative)
            {
                // Negative entries are present. Need to correct.
                assert(len > 1UL);
                const double averageDensity = 1.0/(binWidth*len);
                unsigned long nStepsUp = 1;
                for (; nStepsUp<=len; ++nStepsUp)
                    if (thresholdedIntegral(corrected, len, binWidth,
                                            nStepsUp*averageDensity) < 1.0)
                        break;
                assert(nStepsUp < len);

                double xmin = 0.0;
                double xmax = nStepsUp*averageDensity;
                double shift = 0.0;
                const double eps = 10.0*DBL_EPSILON;
                bool converged = false;
                for (unsigned i=0; i<1000; ++i)
                {
                    if ((xmax - xmin)/xmax <= eps ||
                        (xmin == 0.0 && xmax - xmin < eps))
                    {
                        converged = true;
                        shift = (xmin + xmax)/2.0;
                        break;
                    }
                    const double xtry = (xmin + xmax)/2.0;
                    const double integ = thresholdedIntegral(corrected, len, binWidth, xtry);
                    if (integ == 1.0)
                    {
                        converged = true;
                        shift = xtry;
                        break;
                    }
                    if (integ < 1.0)
                        xmax = xtry;
                    else
                        xmin = xtry;
                }
                if (!converged) throw std::runtime_error(
                    "In npstat::correctDensityEstimateGHU: algorithm did not converge");
                for (unsigned long i=0; i<len; ++i)
                {
                    if (corrected[i] > shift)
                        corrected[i] -= shift;
                    else
                        corrected[i] = 0.0;
                }
                sum = 0.0L;
                for (unsigned long i=0; i<len; ++i)
                    sum += corrected[i];
                const double integ2 = sum*binWidth;
                assert(fabs(integ2 - 1.0) < 1.0e-5);
                for (unsigned long i=0; i<len; ++i)
                    corrected[i] /= integ2;
            }
        }
    }
}
