#ifndef NPSTAT_VARIABLEBANDWIDTHSMOOTH1DWRAP_HH_
#define NPSTAT_VARIABLEBANDWIDTHSMOOTH1DWRAP_HH_

#include <stdexcept>

#include "npstat/wrap/NumpyTypecode.hh"
#include "npstat/stat/variableBandwidthSmooth1D.hh"

namespace npstat {
    template <class Histo>
    PyObject* variableBandwidthSmooth1DWrap(
        const Histo& histo, const AbsDistribution1D& kernel,
        const double bandwidth,
        const double alpha = 0.5,
        const bool increaseBandwidthAtBoundary = false)
    {
        const int typenum = NumpyTypecode<double>::code;

        if (histo.dim() != 1U) throw std::invalid_argument(
            "In npstat::variableBandwidthSmooth1DWrap: "
            "input histogram must be one-dimensional");
        const unsigned long nBins = histo.nBins();
        npy_intp sh[1];
        sh[0] = nBins;

        PyObject* pilot = PyArray_SimpleNew(1, sh, typenum);
        if (!pilot) throw std::runtime_error(
            "In npstat::variableBandwidthSmooth1DWrap: "
            "failed to allocate the pilot array");
        Py_INCREF(pilot);

        PyObject* array = PyArray_SimpleNew(1, sh, typenum);
        if (!array)
        {
            Py_DECREF(pilot);
            throw std::runtime_error(
                "In npstat::weightedVariableBandwidthSmooth1DWrap: "
                "failed to allocate the result array");
        }
        Py_INCREF(array);

        try
        {
            variableBandwidthSmooth1D(histo, (double*)array_data(pilot), nBins,
                                      kernel, bandwidth, alpha,
                                      increaseBandwidthAtBoundary,
                                      (double*)array_data(array), nBins);
        }
        catch (const std::exception& e) 
        {
            Py_DECREF(array);
            Py_DECREF(pilot);
            throw e;
        }

        PyObject* out = Py_BuildValue("(OO)", pilot, array);
        Py_DECREF(array);
        Py_DECREF(pilot);
        return out;
    }

    template <class Histo>
    PyObject* weightedVariableBandwidthSmooth1DWrap(
        const Histo& histo, const double effectiveSampleSize,
        const int symbetaPower, const double bandwidthFactor = 1.0)
    {
        const int typenum = NumpyTypecode<double>::code;

        if (histo.dim() != 1U) throw std::invalid_argument(
            "In npstat::weightedVariableBandwidthSmooth1DWrap: "
            "input histogram must be one-dimensional");
        const unsigned long nBins = histo.nBins();
        npy_intp sh[1];
        sh[0] = nBins;
        PyObject* array = PyArray_SimpleNew(1, sh, typenum);
        if (!array) throw std::runtime_error(
            "In npstat::weightedVariableBandwidthSmooth1DWrap: "
            "failed to allocate the result array");
        Py_INCREF(array);

        double bw = 0.0;
        try
        {
            bw = weightedVariableBandwidthSmooth1D(
                histo, effectiveSampleSize, symbetaPower,
                (double*)array_data(array), nBins, bandwidthFactor);
        }
        catch (const std::exception& e)
        {
            Py_DECREF(array);
            throw e;
        }

        PyObject* out = Py_BuildValue("(Od)", array, bw);
        Py_DECREF(array);
        return out;
    }

    template <class Histo>
    PyObject* simpleVariableBandwidthSmooth1DWrap(
        const Histo& histo, const int symbetaPower,
        const double bandwidthFactor = 1.0)
    {
        const double n = histo.nFillsInRange();
        return weightedVariableBandwidthSmooth1DWrap(
            histo, n, symbetaPower, bandwidthFactor);
    }
}

#endif // NPSTAT_VARIABLEBANDWIDTHSMOOTH1DWRAP_HH_
